\\ set lds=1 for limits of discrete series, i.e. triple weight 0
dwt2Hwt_oddorth(dom_wt, lds=0) = {
  my(res);
  res = vector(2+#dom_wt);
  if(bitxor(lds, #dom_wt%2), res[2] = 1, res[1] = 1);
  if(lds,
    res[1] += 1;
    res[2] += 1;
  );
  for(u=1, #dom_wt,
    res[2+u] = 2*(dom_wt[#dom_wt+1-u]+u);
  );
  return(res);
};

\\ set lds=1 to restrict to limits of discrete series
enum_wts_oddorth(motw, fixed_rk=0, lds=0) = {
  my(dom_wt,i,res,cur,min_rk,max_rk);
  res = List();
  if(fixed_rk<1,
    min_rk = 1;
    max_rk = motw/2-lds;
  ,
    min_rk = fixed_rk-lds;
    max_rk = min(fixed_rk-lds, motw/2);
  );
  for(rk=min_rk, max_rk,
    dom_wt = vector(rk);
    dom_wt[1] = motw/2-rk;
    while(2*(dom_wt[1]+rk) == motw,
      \\ condition epsilon(orth Langlands para)==+1
      if(Mod(vecsum(dom_wt)+lds*(rk+1), 2)==0,
        listput(res, dwt2Hwt_oddorth(dom_wt, lds));
      );
      i = rk;
      dom_wt[i] += 1;
      while(i>1 && dom_wt[i]>dom_wt[i-1],
        dom_wt[i] = 0;
        i -= 1;
        dom_wt[i] += 1;
      );
    );
  );
  return([x | x<-res]);
};

\\ input: dominant weight for SO_odd
dwt2Hwt_sympl(dom_wt) = {
  my(res);
  res = vector(2+#dom_wt);
  for(u=1, #dom_wt,
    res[2+u] = 2*(dom_wt[#dom_wt+1-u]+u)-1;
  );
  return(res);
};

enum_wts_sympl(motw, fixed_rk=0) = {
  my(dom_wt,i,res,cur,min_rk,max_rk);
  res = List();
  if(fixed_rk<1,
    min_rk = 1;
    max_rk = (motw+1)/2;
  ,
    min_rk = fixed_rk;
    max_rk = min(fixed_rk, (motw+1)/2);
  );
  for(rk=min_rk, max_rk,
    dom_wt = vector(rk);
    dom_wt[1] = (motw+1)/2-rk;
    while(2*(dom_wt[1]+rk) == 1+motw,
      listput(res, dwt2Hwt_sympl(dom_wt));
      i = rk;
      dom_wt[i] += 1;
      while(i>1 && dom_wt[i]>dom_wt[i-1],
        dom_wt[i] = 0;
        i -= 1;
        dom_wt[i] += 1;
      );
    );
  );
  return([x | x<-res]);
};

dwt2Hwt_evenorth(dom_wt) = {
  my(res);
  if(dom_wt[#dom_wt]==0,
    res = vector(1+#dom_wt);
    res[1] = 1;
    res[2] = 1;
    for(u=2, #dom_wt,
      res[1+u] = 2*(dom_wt[#dom_wt+1-u]+u-1);
    );
  ,
    res = vector(2+#dom_wt);
    res[3] = 2*dom_wt[#dom_wt];
    for(u=2, #dom_wt,
      res[2+u] = 2*(dom_wt[#dom_wt+1-u]+u-1);
    );
  );
  return(res);
};

enum_wts_evenorth(motw) = {
  my(dom_wt,i,res,cur);
  res = List();
  for(rk=2, 1+motw/2,
    if(Mod(rk,2)==0,
      dom_wt = vector(rk);
      dom_wt[1] = motw/2-rk+1;
      while(2*(dom_wt[1]+rk-1) == motw,
        if(Mod(vecsum(dom_wt), 2)==0,
          listput(res, dwt2Hwt_evenorth(dom_wt));
        );
        i = rk;
        dom_wt[i] += 1;
        while(i>1 && dom_wt[i]>dom_wt[i-1],
          dom_wt[i] = 0;
          i -= 1;
          dom_wt[i] += 1;
        );
      );
    );
  );
  return([x | x<-res]);
};

enum_wts_sing_evenorth(motw, fixed_rk=0) = {
  my(dom_wt,i,res,cur,min_rk,max_rk,wt);
  res = List();
  if(fixed_rk<1,
    min_rk = 1;
    max_rk = motw/2-1;
  ,
    min_rk = fixed_rk-1;
    max_rk = min(fixed_rk-1, motw/2);
  );
  for(rk=min_rk, max_rk,
    if(!(rk%2),
      dom_wt = vector(rk);
      dom_wt[1] = motw/2-rk;
      while(2*(dom_wt[1]+rk) == motw,
        \\ oddorth sets wt[1]=1 and wt[2]=0, and we modify these later
        wt = dwt2Hwt_oddorth(dom_wt);
        \\ condition epsilon(orth Langlands para)==+1
        if(vecsum(dom_wt)%2,
          wt[1] = 0;
          wt[2] = 2;
        ,
          wt[1] = 2;
        );
        listput(res, wt);
        i = rk;
        dom_wt[i] += 1;
        while(i>1 && dom_wt[i]>dom_wt[i-1],
          dom_wt[i] = 0;
          i -= 1;
          dom_wt[i] += 1;
        );
      );
    );
  );
  return([x | x<-res]);
};


enum_wts_orth(motw) = concat(enum_wts_oddorth(motw), enum_wts_evenorth(motw));

