# Compute Euler Phi function of prod_i p_i^{k_i}
# Input: primes_list is the list of p_i's, exponents the list of k_i's
def totient(primes_list, exponents):
    res = 1
    for i in range(len(primes_list)):
        if exponents[i]>0:
            res = res * (primes_list[i]-1)*primes_list[i]^(exponents[i]-1)
    return res

def prodexp(primes_list, exponents):
    res = 1
    for i, p in enumerate(primes_list):
        res *= p^(exponents[i])
    return res

# returns a list of all [m, phi(m)] such that phi(m)<=n
def euler_range(n):
    primes_list = prime_range(2, n+2)
    exponents = [0]*len(primes_list)
    res = []
    while totient(primes_list, exponents)<=n:
        if prodexp(primes_list, exponents)>2:
            res.append([prodexp(primes_list, exponents),
                        totient(primes_list, exponents)])
        exponents[0] += 1
        j=1
        while j<len(primes_list) and totient(primes_list, exponents)>n:
            exponents[j-1]=0
            exponents[j]=exponents[j]+1
            j=j+1
    return res

# rough enumeration of possible conjugacy classes for Sp_n or SO_n (or rather
# characteristic polynomials, so up to outer conjugation for SO_even)
# such a class is represented as a list [[m_1, k_1], ..., [m_r, k_r]] where the
# integers m_i>=1 are increasing, k_i>=1 is even if m_i=2 and
# sum_i k_i phi(m_i) = n
def it_cc_Qbar(n):
    degrees = euler_range(n)
    mults = [0]
    for j in range(1,len(degrees)):
        mults.append(0)
    d=0
    while d<=n:
        numminusone = 0
        while numminusone+d <= n:
            cc = []
            if numminusone+d < n:
                cc.append([1,n-numminusone-d])
            if numminusone>0:
                cc.append([2,numminusone])
            for i, m in enumerate(degrees):
                if mults[i]>0:
                    cc.append([m[0], mults[i]])
            cc.sort(key=lambda x: x[0])
            yield(cc)
            numminusone += 2
        mults[0]=mults[0]+1
        d = d + degrees[0][1]
        j=1
        while j<len(degrees) and d>n:
            d = d - mults[j-1]*degrees[j-1][1]
            mults[j-1] = 0
            mults[j] = mults[j]+1
            d = d + degrees[j][1]
            j += 1
    return

# see Corollary 3.9 in the paper
def imposed_global_discs(cc):
    disc_one = 1
    disc_minus_one = 1
    for m, mult in cc:
        if m<3 or Mod(mult,2)==0:
            continue
        F = factor(m)
        if len(F)==1:
            if F[0][0]==2:
                if Mod(disc_minus_one, 2)==0:
                    disc_minus_one /= 2
                else:
                    disc_minus_one *= 2
                if Mod(disc_one, 2)==0:
                    disc_one /= 2
                else:
                    disc_one *= 2
            else:
                p = F[0][0]
                if Mod(disc_one, p)==0:
                    disc_one /= p
                else:
                    disc_one *= p
        elif len(F)==2 and F[0][0]==2 and F[0][1]==1:
            p = F[1][0]
            if Mod(disc_minus_one, p)==0:
                disc_minus_one /= p
            else:
                disc_minus_one *= p
    return (disc_one, disc_minus_one)

# checks if a given conjugacy class can intersect hyperspecial subgroups at all
# places
def orth_cc_is_hs(cc):
    disc_one, disc_minus_one = imposed_global_discs(cc)
    if disc_one!=1 and cc[0][0]!=1:
        return False
    if disc_minus_one!=1 and cc[0][0]!=2 and (len(cc)<2 or cc[1][0]!=2):
        return False
    return True

def double_tuple(c):
    return tuple(tuple(x) for x in c)

# for Sp_2n
def cc_Qbar_Sp(n):
    return [double_tuple(cc) for cc in it_cc_Qbar(2*n)]

# for SO_n
def relevant_cc_Qbar_SO(n):
    return [double_tuple(cc) for cc in it_cc_Qbar(n) if orth_cc_is_hs(cc)]

