C = ComplexBallField(150)
R = C.base_ring()
Pi = R(pi)

N = 100
# Bernoulli numbers
#bound_B = 100
#B_list = [bernoulli(2*i) for i in range(bound_B+1)]
#rB_list = [R(x) for x in B_list]

max_mw = 63
pc = [\
  [[1,0], 1], [[0,0,11], 1], [[0,0,15], 1], [[0,0,17], 1],\
  [[0,0,19], 1], [[0,0,21], 1], [[0,1,22], 1], [[0,0,23], 2],\
  [[0,0,7,19], 1], [[0,0,5,21], 1], [[0,0,9,21], 1], [[0,0,13,21], 1],\
  [[0,0,7,23], 1], [[0,0,9,23], 1], [[0,0,13,23], 1],\
  [[0,0,5,13,23], 1], [[0,0,3,15,23], 1], [[0,0,7,15,23], 1],\
  [[0,0,5,17,23], 1], [[0,0,9,17,23], 1], [[0,0,3,19,23], 1],\
  [[0,0,11,19,23], 1], [[0,0,3,11,17,21,23], 1],\
  [[0,1,8,16,24], 1], [[0,0,4,10,18,24], 1], [[0,0,2,14,20,24], 1]\
  ]

pc_as_tup = [tuple(wt) for wt, _ in pc]

def dimw(wt):
    return wt[0]+wt[1]+2*(len(wt)-2)

def wt2multvec(p):
    res = matrix(ZZ, max_mw+2, 1)
    for i, x in enumerate(p):
        if i<2:
            res[i, 0] = x
        else:
            res[1+x, 0] += 1
    return res

def multvec2wt_even(v):
    res = v[:2]
    for i, m in enumerate(v[2:]):
        res += [2+2*i]*m
    return tuple(res)

def multvec2wt_odd(v):
    res = [0,0]
    for i, m in enumerate(v):
        res += [1+2*i]*m
    return tuple(res)

pc_mat = block_matrix(1, len(pc), [wt2multvec(p) for p, _ in pc])
pc_mult = [m for _, m in pc]

tmp_mat = matrix(Integers(4), max_mw+2, max_mw+2)
tmp_mat[0,1] = tmp_mat[1,0] = Mod(1,4)
for w in range(1, max_mw+1):
    tmp_mat[0, 1+w] = tmp_mat[1, 1+w] = tmp_mat[1+w, 0] = tmp_mat[1+w, 1]\
            = Mod(1+w, 4)
    tmp_mat[1+w, 1+w] = Mod(2+2*w, 4)
    for w2 in range(w+1, max_mw+1):
        tmp_mat[1+w, 1+w2] = tmp_mat[1+w2, 1+w] = Mod(2+2*w2, 4)
epsil_mat = pc_mat.transpose()*tmp_mat

def phi(x):
    return 1/2*psi((C(x)+1)/2) - 1/2*psi(C(x)/2)

def r(x):
    return 2*Pi^2*exp(-x)/(x^2+Pi^2)^2

def psi_prime(z):
    return C(z).psi(1)

def phi_prime(x):
    return 1/4*psi_prime((C(x)+1)/2) - 1/4*psi_prime(C(x)/2)

load("func_noGRH.sage")

def triv(l):
    return (J(0,l) + diff1eps(l))/2

def eps(l):
    return (J(0,l) - diff1eps(l))/2

def update_cache(l):
    # triv, eps, I_1, ..., I_{max_mw}
    biloo = matrix(R, max_mw+2, max_mw+2)
    phi0l = Phi0(l)
    ca_J = [J(ZZ(w), l)/phi0l for w in range(2*max_mw+1)]
    biloo[1,1] = biloo[0,0] = triv(l)/phi0l
    biloo[1,0] = biloo[0,1] = eps(l)/phi0l
    for w in range(1, max_mw+1):
        biloo[0, 1+w] = biloo[1, 1+w] = biloo[1+w, 0] = biloo[1+w, 1] = ca_J[w]
        biloo[1+w,1+w] = ca_J[2*w] + ca_J[0]
        for w2 in range(w+1, max_mw+1):
            biloo[1+w, 1+w2] = biloo[1+w2, 1+w] = ca_J[w+w2] + ca_J[w2-w]
    G = diagonal_matrix(R, [R(1/m) for _, m in pc])\
            - pc_mat.transpose()*biloo*pc_mat
    epsil_pc_mat = epsil_mat*pc_mat
    e_orth = Phi12(l)/phi0l
    for i in range(len(pc)):
        for j in range(i+1, len(pc)):
            if epsil_pc_mat[i,j]==Mod(2,4):
                G[i,j] -= e_orth/2
                G[j,i] = G[i,j]
    return [biloo, G, -pc_mat.transpose()*biloo, e_orth]

# for the parameter l (lambda) we only need so much precision
def float2rat(l):
    return floor(10000*l+0.5)/10000

def gramoo_lower_oddw(w, l0, N):
    l = float2rat(l0)
    if not l in ca:
        ca[l] = update_cache(l)
    biloo, G, U, e_orth = ca[l]
    n = ZZ((w+1)/2)
    res = matrix(ZZ, n, n)
    for i in range(n):
        res[i, i] = (biloo[2*i+2, 2*i+2]*N).lower().floor()
        for j in range(i+1, n):
            res[i, j] = (biloo[2*i+2, 2*j+2]*N).lower().floor()
            res[j, i] = res[i, j]
    return res

def gramoo_lower_evenw(w, l0, N):
    l = float2rat(l0)
    if not l in ca:
        ca[l] = update_cache(l)
    biloo, G, U, e_orth = ca[l]
    n = ZZ(w/2+2)
    res = matrix(ZZ, n, n)
    res[1,1] = res[0,0] = (biloo[0,0]*N).lower().floor()
    res[1,0] = res[0,1] = (biloo[1,0]*N).lower().floor()
    for i in range(2,n):
        res[0, i] = res[1, i] = res[i, 0] = res[i, 1] =\
                (biloo[0, 2*i-1]*N).lower().floor()
        res[i, i] = (biloo[2*i-1, 2*i-1]*N).lower().floor()
        for j in range(i+1, n):
            res[i, j] = (biloo[2*i-1, 2*j-1]*N).lower().floor()
            res[j, i] = res[i, j]
    return res

def mask2list(m):
    res = []
    mi = m
    i = 0
    while mi!=0:
        if Mod(mi, 2)==0:
            mi /= 2
        else:
            res.append(i)
            mi = (mi-1)/2
        i += 1
    return res

def CF_over_Phi0(db, wt, mult):
    if not (wt in db):
        raise ValueError(str(wt)+' not in the database')
    l0, min_eig, mask, vec = db[wt]
    l = float2rat(l0)
    if not l in ca:
        ca[l] = update_cache(l)
    biloo, G, U, e_orth = ca[l]
    p = wt2multvec(wt)
    ml = mask2list(mask)
    v = matrix(R, len(pc), 1)
    epsil = epsil_mat*p
    res = R(0)
    v_last = R(vec[-1])
    for i, j in enumerate(ml):
        v[j, 0] = R(vec[i])
        if mult==1 and epsil[j, 0]==Mod(2,4):
            res += -e_orth*v[j,0]*v_last
    res += v_last^2*(1/mult-(p.transpose()*biloo*p)[0,0])
    res += 2*v_last*(v.transpose()*U*p)[0,0]
    res += (v.transpose()*G*v)[0,0]
    err = (res-R(min_eig)).abs().upper()
    #if err>0.00001:
    #    raise ValueError,\
    #        "Precision error:"+str(wt)+", "+str(res)+", "+str(err)
    return [res, err]

def CF_over_Phi0_pair_nodb(wt1, mult1, wt2, mult2, l0, min_eig, t1, t2):
    l = float2rat(l0)
    if not l in ca:
        ca[l] = update_cache(l)
    biloo = ca[l][0]
    p1 = wt2multvec(wt1)
    p2 = wt2multvec(wt2)
    v1 = R(t1)
    v2 = R(t2)
    res = R(0)
    res += v1^2*(1/mult1-(p1.transpose()*biloo*p1)[0,0])
    res += v2^2*(1/mult2-(p2.transpose()*biloo*p2)[0,0])
    res -= 2*v1*v2*(p2.transpose()*biloo*p1)[0,0]
    err = (res-R(min_eig)).abs().upper()
    return [res, err]

# recursive helper function for short_effective_vectors below
def sev_rec(x, n, i, L, D, bound):
    if i<0:
        if bound>=0:
            yield x
        return
    # the floor() below are exact (see Sage documentation)
    lb = max(0, -floor(sqrt(bound/D[i])
                + sum(L[j,i]*x[j-i-1] for j in range(i+1, n))))
    ub = floor(sqrt(bound/D[i])
               - sum(L[j,i]*x[j-i-1] for j in range(i+1, n)))
    for xi in range(lb, ub+1):
        new_bound = bound\
                    - D[i]*(xi + sum(L[j,i]*x[j-i-1] for j in range(i+1, n)))^2
        for res in sev_rec([xi]+x, n, i-1, L, D, new_bound):
            yield res
    return

# qfminim using the first (simpler) algorithm in Fincke-Pohst, restricting to
# effective vectors
# assumption: G is a square positive definite matrix with rational coefficients
# of dimension>0
def short_effective_vectors(G, bound):
    L, D = G.indefinite_factorization()
    # now G = L * D * L.transpose() with D diagonal and L lower triangular unipotent
    n = G.nrows()
    for res in sev_rec([], n, n-1, L, D, bound):
        yield res
    return

def enum_possible_pi(w, l, mult, N):
    if Mod(w,2)==0:
        G = gramoo_lower_evenw(w, l, N)
    else:
        G = gramoo_lower_oddw(w, l, N)
    for res in short_effective_vectors(G, N/mult):
        if res[-1]>0 and (Mod(w-1, 2)==0 or Mod(sum(res[1:]), 2)==0):
            yield res
    return

# verify, for each weight wt in L, whether the vector in db[wt] gives a negative
# C^F value
# all elements of L must have a motivic weight having same parity as w
def verify_enum(db, w, L, mult, verbose=False):
    count = 0
    count_bad = 0
    max_err = 0
    list_bad = []
    if Mod(w,2)==0:
        multvec2wt = multvec2wt_even
    else:
        multvec2wt = multvec2wt_odd
    for wt in L:
        count += 1
        if mult<2 and wt in pc_as_tup:
            continue
        res, err = CF_over_Phi0(db, wt, mult)
        max_err = max(max_err, err)
        if not(res<0):
            list_bad.append([wt, res])
            if not(res>=0):
                print(str(wt)+", ambiguity: "+str(res))
            count_bad += 1
            if verbose:
                print(wt)
            #last_coeff = R(db[wt][3][-1])
            #inv_mult_bound = -res/last_coeff^2 + 1/mult
            #if verbose:
            #    if mult<2 or not(inv_mult_bound>0):
            #        print(str(wt))
            #    else:
            #        mult_bound = (1/inv_mult_bound).upper().floor()
            #        print(str(wt)+", multiplicity <= "+str(mult_bound))
    print(count_bad, "possible pi_infty remaining,", count, "initially.")
    print(" maximal difference with precomputed value: "+str(max_err))
    return list_bad

def verify(db, w, l, mult, N, verbose=False):
    count = 0
    count_bad = 0
    L = []
    if Mod(w,2)==0:
        multvec2wt = multvec2wt_even
    else:
        multvec2wt = multvec2wt_odd
    return verify_enum(db, w, map(multvec2wt, enum_possible_pi(w, l, mult, N)),
                       mult, verbose)

def verify_reg_selfdual(db, w, d, verbose=True, ret_bad=False):
    count_bad = 0
    count_non_ex = 0
    max_err = 0
    L = []
    if Mod(w,2)==0:
        multvec2wt = multvec2wt_even
        if Mod(d,2)==0:
            wt2dw = wt2dw_SO_even
        else:
            wt2dw = wt2dw_Sp
    else:
        multvec2wt = multvec2wt_odd
        wt2dw = wt2dw_SO_odd
    for wt in db:
        if wt[-1]!=w or dimw(wt)!=d:
            continue
        dw = wt2dw(wt)
        res, err = CF_over_Phi0(db, wt, 1)
        max_err = max(max_err, err)
        if res<0:
            count_non_ex += 1
            if not ret_bad:
                L.append(dw)
        else:
            if ret_bad:
                L.append(wt)
            count_bad += 1
    if verbose:
        print("Dimension "+str(d)+", weight "+str(w)+": "\
                +str(count_non_ex)+" non-existing, "+str(count_bad)+" unknown")
        print(" maximal difference with precomputed value:"+str(max_err))
    return L
